#include "woody.h"

int wdy_get_file(char *path, t_map *file) {
	int fd = open(path, O_RDONLY);
	if (fd == -1) {
		return wdy_perror(path);
	}

	file->size = lseek(fd, 0, SEEK_END);
	if (file->size == -1) {
		close(fd);
		return wdy_perror(path);
	}

	file->data = mmap(
			NULL,
			file->size,
			PROT_READ | PROT_WRITE,
			MAP_PRIVATE,
			fd,
			0);
	close(fd);
	if (file->data == MAP_FAILED) {
		return wdy_perror(path);
	}

	return RET_OK;
}

int pack_elf32(t_map file) {
	(void)file;
	return wdy_error("ELF needs to be in 64 bits format");
}


int get_first_load_segment64(
		Elf64_Ehdr elf_header,
		Elf64_Phdr *program_headers,
		Elf64_Phdr **first_load_segment) {
	for (int i = 0; i < elf_header.e_phnum; i++) {
		Elf64_Phdr *p_hdr = &program_headers[i];
		if (p_hdr->p_type == PT_LOAD) {
			*first_load_segment = p_hdr;
			return RET_OK;
		}
	}
	return RET_ERR;
}
int get_load_segment64(
		Elf64_Ehdr elf_header,
		Elf64_Phdr *program_headers,
		Elf64_Phdr **load_segment) {
	for (int i = 0; i < elf_header.e_phnum; i++) {
		Elf64_Phdr *p_hdr = &program_headers[i];
		if (p_hdr->p_type == PT_LOAD
				&& p_hdr->p_flags & PF_X
				&& p_hdr->p_vaddr <= elf_header.e_entry
				&& p_hdr->p_vaddr + p_hdr->p_filesz > elf_header.e_entry) {
			*load_segment = p_hdr;
			return RET_OK;
		}
	}
	return RET_ERR;
}

int get_code_cave64(t_map file, Elf64_Phdr load_segment, t_map *code_cave) {
	size_t page_size = load_segment.p_align;
	size_t len = page_size - load_segment.p_filesz % page_size; 
	size_t offset = load_segment.p_offset + load_segment.p_filesz;
	unsigned char *seg = fetch(file, offset, len);
	if (!seg) {
		return RET_ERR;
	}

	size_t longest = 0;
	size_t longest_i = 0;
	for (size_t i = 0; i < len; i++) {
		if (seg[i] == 0) {
			size_t j = 0;
			while (i + j < len && seg[i + j] == 0) {
				j++;
			}
			if (j > longest) {
				longest_i = i;
				longest = j;
			}
			i += j;
		}
	}

	code_cave->data = fetch(file, longest_i + offset, longest);
	code_cave->size = longest;
	return RET_OK;
}

int pack_elf64(t_map file) {
	Elf64_Ehdr *elf_header = (Elf64_Ehdr *)fetch(file, 0, sizeof(Elf64_Ehdr));
	if (!elf_header) {
		return wdy_error("cannot fetch elf header");
	}

	size_t phdrs_len = elf_header->e_phnum * elf_header->e_phentsize;
	Elf64_Phdr *program_headers = fetch(file, elf_header->e_phoff, phdrs_len);
	if (!program_headers) {
		return wdy_error("cannot fetch program headers table");
	}

	Elf64_Phdr *first_load_segment;
	if (get_first_load_segment64(*elf_header, program_headers, &first_load_segment) == RET_ERR) {
		return wdy_error("cannot get first load segment");
	}

	Elf64_Phdr *load_segment;
	if (get_load_segment64(*elf_header, program_headers, &load_segment) == RET_ERR) {
		return wdy_error("cannot get load segment");
	}

	// TODO the whole section below till the end of the function needs 
	// cleaning and refactoring, better naming also.

	// This should go in a separate function regarding encryption if encryption has been requested.
	/*
	   size_t encryption_block_size = 31; // 32 bits and padding...
	   t_map to_encrypt;

	   to_encrypt.data = fetch(file, load_segment.p_offset, 0);
	// size is a multiple of 31...
	to_encrypt.size = load_segment.p_filesz + encryption_block_size - load_segment.p_filesz % encryption_block_size;
	 */

	t_map code_cave;
	if (get_code_cave64(file, *load_segment, &code_cave) == RET_ERR) {
		return wdy_error("can't get code cave");
	}

	t_payload64 payload = get_xor_payload64();
	ft_printf("info: using %s algorithm\n", payload.algo_name);
	free(payload.algo_name);
	// This should fallback to compression algorithm, or smaller payload (eg rsa->xor)
	if (payload.len > (size_t)code_cave.size) {
		printf("code cave size: %ld (0x%lx) bytes\n", code_cave.size, code_cave.size);
		return wdy_error("payload length exceed code cave size");
	}

	// e_entry because relative to this (where we gonna go)
	// code cave start because thats the start of the code
	// jump_offset is the index of jump from code cave start
	// 4 because jump has a 4 byte operand)
	size_t code_cave_start = code_cave.data - file.data;
	int jump_value = elf_header->e_entry - first_load_segment->p_paddr - code_cave_start - payload.jump_offset - 4;
	ft_memcpy(payload.data + payload.jump_offset, &jump_value, sizeof(jump_value));

	if (payload.encrypt != NULL) {
		uint64_t load_ptr_value = code_cave_start - load_segment->p_offset;
		ft_memcpy(payload.data + payload.load_ptr_offset, &load_ptr_value, sizeof(load_ptr_value));
		ft_memcpy(payload.data + payload.load_size_offset, &load_segment->p_filesz, sizeof(load_segment->p_memsz));
		void *key = NULL;
		if (payload.gen_key(&key) != RET_OK) {
			return wdy_error("key generation failed");
		}
		ft_memcpy(payload.data + payload.private_key_offset, key, sizeof(uint64_t));
		payload.encrypt(file, key, *load_segment);
		free(key);
	}
	elf_header->e_entry = code_cave.data - file.data + first_load_segment->p_paddr;
	load_segment->p_filesz += payload.len;
	load_segment->p_memsz += payload.len;
	load_segment->p_flags |= PF_W | PF_R;
	ft_memcpy(code_cave.data, payload.data, payload.len);
	free(payload.data);

	int fd = open("woody", O_WRONLY | O_CREAT, 0755);
	if (fd == -1) {
		return wdy_perror("woody");
	}
	write(fd, file.data, file.size);
	munmap(file.data, file.size);
	return RET_OK;
}

int pack_elf(t_map file) {	
	// at this point arch is known to be either ELFCLASS32 or ELFCLASS64
	unsigned char arch = ((unsigned char *)file.data)[4];
	if (arch == ELFCLASS64) {
		return pack_elf64(file);
	}
	return pack_elf32(file);
}

int woody_woodpacker(char *path) {
	t_map file;

	if (wdy_get_file(path, &file) == RET_ERR) {
		return RET_ERR;
	}
	if (check_ident(file) == RET_ERR) {
		return RET_ERR;
	}
	int ret = pack_elf(file);
	if (ret == RET_OK) {
		ft_printf("successfully woody woodpacked %s\n", path);
	} else {
		ft_printf("error: can't woody woodpack %s\n", path);
	}
	return ret;
}
